from datetime import datetime

# upper, lower band (Bollinger)
K_ARR = [0.5, 1, 1.5, 2]

#---------------------------------------------------------------------------------------------------------------------
# Peer Analysis
#---------------------------------------------------------------------------------------------------------------------
# Analysis
METHODS = ['Ratio', 'zscore', 'DROsMA'] #['ratio', 'ROsMA', 'OLS', 'TLS', 'KF', 'NPI']
FROM_ = "2015-01-01" #datetime.fromordinal(datetime(1900, 1, 1).toordinal() + int(worksheet.cell_value(1, 1)) - 2).strftime("%Y-%m-%d")#"2015-01-01"#"2015-01-01"

TO_ = datetime.now().strftime("%Y-%m-%d") #datetime.fromordinal(datetime(1900, 1, 1).toordinal() + int(worksheet.cell_value(2, 1)) - 2).strftime("%Y-%m-%d")
ROLLING = True
WINDOW = 365#int(worksheet.cell_value(3, 1))

ADDITIONAL_FEATURES = ['AVG_TRADING_VALUE', 'DIVIDEND_YIELD']
#---------------------------------------------------------------------------------------------------------------------
# Exporting
#-------------------key, name, width, format, description--------------------------------------------------------------
PLOT_FROM_ = FROM_
PLOT_XAXIS_N_LOCATOR = 10

SHEET_HEIGHT = 210
SHEET_RECIPES = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 13, 14, 15, 16, 17, 18, 19, 20]
SHEET_VARIABLES = [['LONG', 'Y'], ['SHORT', 'X']]#[['COMM', '보통주'], ['PREF', '우선주']]#[['X', '독립변수'], ['Y', '종속변수']] #

SHEET_COOKBOOK  =   {0:["NO.", 6.38, "numeric", "순서"],
                     1:["SYMBOL", 15, "string", "종목코드"],
                     2:["SYMBOL", 15, "string", "종목코드"],
                     3:["NAME", 15, "string", "종목명"],
                     4:["NAME", 15, "string", "종목명"],

                     5:["SPREAD", 13.38, "float_third_dec", "스프레드"],

                     6:["PRICE_PLOT", 101, "string", "가격 그래프"],
                     7:["SPREAD_PLOT", 101, "string", "스프레드 그래프"],
                     8:["BOX_PLOT", 40, "string", "스프레드 박스 그래프"],
                     9:["DIST_PLOT", 40, "string", "스프레드 분포 그래프"],
                     10:["LOG_TLS_PLOT", 40, "string", "로그 주가 Total Least Squares 회귀 그래프"],

                     11:["CLOSE", 15, "string", "심볼"],
                     12:["CLOSE", 15, "string", "심볼"],

                     13:["CORRELATION", 15, "string", "심볼"],

                     14:["P_VALUE", 15, "float_second_dec", "스프레드의 통계값이 실제로 관측된 값 이상일 확률"
                                                            "(0.05 이하)"],
                     15:["HURST_EXPNT", 15, "float_second_dec", "스프레드 허스트 지수"],
                     16:["HALF_LIFE", 15, "float_second_dec", "스프레드 반감기"],

                     17:["AVG_TRADING_VALUE", 21, "integer", "30일 평균 거래대금"],
                     18:["AVG_TRADING_VALUE", 21, "integer", "30일 평균 거래대금"],
                     19:["DIVIDEND_YIELD", 16.63, "float_rate", "배당 수익률(12M)"],
                     20:["DIVIDEND_YIELD", 16.63, "float_rate", "배당 수익률(12M)"]}

def cook_sheet(cookbook, recipes, variables):
    sheet_columns = {}
    x_column, y_column = variables[0][0], variables[1][0]
    x_description, y_description = variables[0][1], variables[1][1]

    for recipe in recipes:
        target = cookbook[recipe]
        if recipe in [1, 3, 11, 17, 19]:
            target[0] = x_column + '_' + target[0]
            target[3] = x_description + ' ' + target[3]
        elif recipe in [2, 4, 12, 18, 20]:
            target[0] = y_column + '_' + target[0]
            target[3] = y_description + ' ' + target[3]

        sheet_columns[recipe] = target
    return sheet_columns

SHEET_COLUMNS = cook_sheet(SHEET_COOKBOOK, SHEET_RECIPES, SHEET_VARIABLES)




from yahoofinancials import YahooFinancials
import pandas as pd
from datetime import datetime
from utils import find_key_by_val

class YahooStockChart():
    def __init__(self):
        self.fields_dic = {"DATE": "formatted_date", "CLOSE": "adjclose", "OPEN": "open", "HIGH": "high", "LOW": "low",
                           "VOLUME": "volume"}

        self.freq_dic = {"1d": "daily", "1w": "weekly", "1m": "monethly"}

    def request(self, symbol, from_date='2015-01-01', to_date=datetime.now().strftime("%Y-%m-%d"), fields=["OPEN", "HIGH", "LOW", "CLOSE", "VOLUME"], freq="1d"):

        req_fields = ["formatted_date"] + [self.fields_dic[field] for field in fields]

        yahoo_financials = YahooFinancials([symbol])
        dataset_raw = yahoo_financials.get_historical_price_data(start_date=from_date,
                                                                 end_date=datetime.now().strftime("%Y-%m-%d"),
                                                                 time_interval='daily')
        dataset = pd.DataFrame.from_dict(dataset_raw[symbol]['prices'])
        dataset = dataset[req_fields]
        dataset.set_index('formatted_date', drop=True, inplace=True)
        dataset.index = pd.to_datetime(dataset.index)
        dataset.index.name = None
        dataset.columns = fields
        for column in dataset.columns:
            dataset[column] = pd.to_numeric(dataset[column])

        start_date = datetime.strptime(from_date, "%Y-%m-%d")
        dataset = dataset.loc[(start_date <= dataset.index) & (dataset.index <=to_date), fields]
        return dataset


class YahooEvents:
    def __init__(self):
        self.freq_dic = {
            "1d": "daily",
            "1w": "weekly",
            "1m": "monethly"
        }
        self.fields_dic = {
            'DIVIDENDS': 'dividends'
        }

    def request(self, symbol, from_date="1990-01-01", to_date=datetime.now().strftime("%Y-%m-%d"),
                field_arr=["DIVIDENDS"], freq="1d"):
        yahoo_financials = YahooFinancials([symbol])
        events_raw = \
            yahoo_financials.get_historical_price_data(start_date=from_date,
                                                       end_date=to_date,
                                                       time_interval=self.freq_dic[freq])[symbol]['eventsData']

        req_field_arr = [self.fields_dic[field] for field in field_arr]
        res = {}
        for req_field in req_field_arr:
            # res_dic = events_raw[req_field]
            # res[] = {k: v['amount'] for k, v in res_dic.items()}
            dividends = pd.DataFrame.from_dict(events_raw[req_field]).T

            dividends.set_index('formatted_date', drop=True, inplace=True)
            dividends = dividends[['amount']]

            dividends.columns = ['DIVIDENDS']
            dividends.index.name = None
            dividends.index = pd.to_datetime(dividends.index)
            dividends = dividends.sort_index()
            dividends['DIVIDENDS'] = pd.to_numeric(dividends['DIVIDENDS'])

            res[find_key_by_val(self.fields_dic, req_field)] = dividends
        return res


def main():
    yahoo = YahooStockChart()
    print(yahoo.request('CEM'))

if __name__ == '__main__':
    main()


import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.font_manager as fm
import os
from matplotlib.ticker import MaxNLocator
import seaborn as sns
import numpy as np
from SETTINGS import K_ARR, PLOT_XAXIS_N_LOCATOR
import matplotlib.dates as mdates
from utils import split_yearly

# Constants
FONT_SIZE = 8
LINE_WIDTH = 0.6
GRID_ALPHA = 0.3
DPI = 300
FIG_HEIGHT = 2.5

# Static Fonts 추가
font_dirs = [os.path.join(os.path.dirname(os.path.dirname(__file__)), 'fonts'), ]
font_files = fm.findSystemFonts(fontpaths=font_dirs, fontext='ttf')
font_list = fm.createFontList(font_files)
fm.fontManager.ttflist.extend(font_list)

# Default Settings
mpl.rcParams['figure.dpi'] = mpl.rcParams['savefig.dpi'] = DPI
mpl.rcParams['font.size'] = mpl.rcParams['legend.fontsize'] = mpl.rcParams['figure.titlesize'] = FONT_SIZE
mpl.rcParams['lines.linewidth'] = LINE_WIDTH
mpl.rcParams['axes.unicode_minus'] = False

mpl.rcParams['grid.alpha'] = GRID_ALPHA
mpl.rcParams['axes.grid'] = True

mpl.rcParams['xtick.labelsize'] = FONT_SIZE
mpl.rcParams['ytick.labelsize'] = FONT_SIZE
# Font
font_name = [(f.name, f.fname) for f in fm.fontManager.ttflist if 'Nanum' in f.name][0]
mpl.rc('font', family=font_name)
font = fm.FontProperties(family=font_name, size=FONT_SIZE)

method_category = {'diff': 'index',
                   'Ratio': 'index',
                   'ROsMA': 'spread',
                   'DROsMA': 'spread',
                   'OLS': 'spread',
                   'TLS': 'spread',
                   'KF': 'spread',
                   'NPI': 'spread',
                   'zscore': 'spread'}

class PairPlotLib():
    def __init__(self, pair, from_='2015-01-01', to_='2020-12-31'):
        self.pair = pair
        self.x_name = pair.x_asset.name
        self.y_name = pair.y_asset.name
        self.from_, self.to_ = from_, to_

    @staticmethod
    def plot_yearly(df, normalize=False, scale=False):
        yearly_df = split_yearly(df)
        fig = plt.figure(figsize=(7, FIG_HEIGHT), dpi=DPI)
        ax = fig.add_subplot(111)

        color_pallette = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f',
                          '#bcbd22', '#17becf']
        years = yearly_df.columns

        for year, color in zip(years, color_pallette):
            if year == years[0]:  # 첫 해 제거
                continue
            data = yearly_df[year]

            if normalize:
                std = np.std(data)
                mean = np.mean(data)
                data = (data - mean) / std

            if scale:  # 스케일링이 안 됨
                data = data / data[data.first_valid_index()]

            ax.plot(data, label=year, color=color)

            bbox_props = dict(boxstyle="larrow", pad=0.01, fc=color, ec="b", lw=0)
            plt.annotate(year + ': %.3f   ' % data[-1], xy=(1, data[-1]), xytext=(0, 0),
                         xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                         bbox=bbox_props, color='white')

        ax.set_ylabel('Yearly Seasonality')
        ax.tick_params(bottom=False, labelbottom=True)
        #ax.tick_params(labelleft=False, labelright=True, left=False, right=True)

        fig.tight_layout()

        every_nth = 30
        for n, label in enumerate(ax.xaxis.get_ticklabels()):
            if n % every_nth != 0:
                label.set_visible(False)

        ax.xaxis.grid(False, which='both')
        plt.close(fig='all')
        return fig, ax, yearly_df


    @staticmethod
    def plot(x, title, label, from_='2015-01-01', to_='2020-12-31'):
        fig = plt.figure(figsize=(7, FIG_HEIGHT), dpi=DPI)
        ax = fig.add_subplot(111)
        x = x.loc[from_:to_]

        ax.plot(x, label=label, color='blue')

        ax.legend(loc='upper left')

        ax.set_ylabel(title)
        ax.tick_params(labelleft=False, labelright=True, left=False, right=True)

        fig.tight_layout()
        annotations = [[x, "C:", "blue"]]
        for annotation in annotations:
            data, prefix, box_color = annotation[0], annotation[1], annotation[2]
            if len(data) == 0:
                break
            bbox_props = dict(boxstyle="larrow", pad=0.01, fc=box_color, ec="b", lw=0)
            plt.annotate(prefix + '%.3f   ' % data[-1], xy=(1, data[-1]), xytext=(0, 0),
                         xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                         bbox=bbox_props, color='white')

        plt.close(fig='all')
        return fig, ax

    def price(self, x, y, title='Price'):
        fig = plt.figure(figsize=(7, FIG_HEIGHT), dpi=DPI)
        ax = fig.add_subplot(111)
        x = x.loc[self.from_:self.to_]
        y = y.loc[self.from_:self.to_]

        ax.plot(x, label=self.x_name, color='red')
        ax.plot(y, label=self.y_name, color='blue')
        ax.legend(loc='upper left')

        ax.set_ylabel(title)
        ax.tick_params(labelleft=False, labelright=True, left=False, right=True)

        fig.tight_layout()
        annotations = [[x, "C:", "red"], [y, "C:", "blue"]]
        for annotation in annotations:
            data, prefix, box_color = annotation[0], annotation[1], annotation[2]
            if len(data) == 0:
                break
            bbox_props = dict(boxstyle="larrow", pad=0.01, fc=box_color, ec="b", lw=0)
            plt.annotate(prefix + '%.3f   '% data[-1], xy=(1, data[-1]), xytext=(0, 0),
                         xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                         bbox=bbox_props, color='white')

        plt.close(fig='all')
        return fig, ax

    def spread_dist(self, method='ROsMA', ylabel='Spread'):
        figsize_x = 2.7
        figsize_y = FIG_HEIGHT

        fig = plt.figure(figsize=(figsize_x, figsize_y), dpi=DPI)
        ax = fig.add_subplot(111)

        spread = self.pair.spread[method]
        if len(spread) > 0:
            sns.distplot(np.array(spread), color='#539caf', label='{} Distribution'.format(method), ax=ax)
            # sns.kdeplot(tuple(np.array(spread)), shade=True, cut=0, color='#539caf', label='{} Spread Density'.format(method), ax=ax)
            sns.rugplot(np.array(spread), color='#539caf', ax=ax)

            ax.legend(loc='upper left')
            ax.axvline(x=spread[-1], color='red', linestyle='--')
            ax.set_ylabel(ylabel)

        plt.tight_layout()
        plt.close(fig='all')
        return fig, ax

    @staticmethod
    def annotate_divs(ax, dvd_dic, method, color='green', fontsize=6, div_yield=False, xytext=(0,10)):

        for x, value in dvd_dic.items():
            nearest = value['nearest']
            if div_yield == True:
                annotation = '{0:.2%}'.format(value['dividend_yield'])
            else:
                annotation = '{0:.2f}'.format(value['dividend_yield'])

            if method == 'Price':
                y = value['price']
            else:
                y = value[method]

            ax.annotate(annotation,
                        (mdates.date2num(nearest), y),
                        xytext=xytext,
                        textcoords='offset points',
                        arrowprops=dict(arrowstyle='-|>'),
                        ha='center',
                        fontsize=fontsize,
                        color=color)


        return ax

    def spread_box(self, method='ROsMA', y_label='Spread'):
        figsize_x = 2.7
        figsize_y = FIG_HEIGHT

        fig = plt.figure(figsize=(figsize_x, figsize_y), dpi=DPI)
        ax = fig.add_subplot(111)

        spread = self.pair.spread[method]

        if len(spread) > 0:
            ax.boxplot(spread)

            if len(self.pair.data[method + '_mean']) > 0:

                for i, k in enumerate(K_ARR):
                    ub_suffix = '_ub_' + str(k)
                    lb_suffix = '_lb_' + str(k)

                    ub_y = self.pair.data[method + ub_suffix][-1]
                    lb_y = self.pair.data[method + lb_suffix][-1]

                    ub_prefix = str(k).rjust(5) + 'σ :'
                    lb_prefix = str(-k).rjust(5) + 'σ :'

                    ub_bbox_props = dict(boxstyle='larrow', pad=0.01, fc='gray', ec="b", lw=0,
                                         alpha=1 / (len(K_ARR) + 1) * (i + 2))
                    lb_bbox_props = dict(boxstyle='larrow', pad=0.01, fc='#539caf', ec="b", lw=0,
                                         alpha=1 / (len(K_ARR) + 1) * (i + 2))

                    ub_text = ub_prefix + '%.3f' % ub_y
                    ax.annotate(ub_text.rjust(15), xy=(1, ub_y), xytext=(0, 0),
                                xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                                bbox=ub_bbox_props, color='white')

                    lb_text = lb_prefix + '%.3f' % lb_y
                    ax.annotate(lb_text.rjust(15), xy=(1, lb_y), xytext=(0, 0),
                                xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                                bbox=lb_bbox_props, color='white')

            ax.axhline(y=spread[-1], color='blue', linestyle='--')

        ax.set_ylabel(y_label)
        ax.set_xlabel(method)

        ax.tick_params(labelbottom=False)



        plt.tight_layout()
        plt.close(fig='all')
        return fig, ax

    def linear_regression(self, method='TLS'):
        log_x = self.pair.data['log_x']
        log_y = self.pair.data['log_y']
        rand_x = np.linspace(log_x.min(), log_x.max(), 100)

        figsize_x = 2.7
        figsize_y = FIG_HEIGHT

        fig = plt.figure(figsize=(figsize_x, figsize_y), dpi=DPI)
        ax = fig.add_subplot(111)

        slope = self.pair.hedge_ratio[method]
        intercept = self.pair.intercept[method]

        ax.scatter(x=log_x, y=log_y, c='black', alpha=0.5, linewidth=1, linestyle='-.', s=5, label='Observed')
        ax.plot(rand_x, rand_x * slope + intercept, color='red', label='y=%.3fx%+.3f' % (slope, intercept))

        ax.legend(loc='upper left')
        ax.set_ylabel(self.y_name)
        ax.set_xlabel(self.x_name)

        plt.tight_layout()
        plt.close(fig='all')
        return fig, ax

    @staticmethod
    def spread_staticmethod(spread_arr, methods, window=365, method_category='index', y_label='Spread', bollinger=True):
        n_fig = len(methods)
        figsize_x = 7
        figsize_y = n_fig * FIG_HEIGHT

        fig, axes = plt.subplots(nrows=n_fig, ncols=1, figsize=(figsize_x, figsize_y), dpi=DPI)

        axes = [axes, None] if len(methods) == 1 else axes  # padding for zip function

        for spread, method, ax in zip(spread_arr, methods, axes):
            if len(spread) > 0:
                ax.plot(spread, label=method, color='blue')

                if method_category == 'total_mean':
                    mean = spread.mean()
                    ax.axhline(mean, linestyle='--', color='red')
                elif method_category == 'spread':
                    ax.axhline(0, linestyle='--', color='red')
                else:
                    pass

                if bollinger is True:
                    rolling_spread = spread.rolling(window='{}D'.format(window), min_periods=int(window / 2))
                    rolling_mean = rolling_spread.mean().dropna()
                    rolling_std = rolling_spread.std().dropna()

                    ub_arr = {}
                    lb_arr = {}

                    for k in K_ARR:
                        ub_suffix = '_ub_' + str(k)
                        lb_suffix = '_lb_' + str(k)
                        ub_arr[method + ub_suffix] = rolling_mean + k * rolling_std
                        lb_arr[method + lb_suffix] = rolling_mean - k * rolling_std

                    rolling_mean = rolling_mean.dropna()
                    # mean
                    if len(rolling_mean) > 0 and method_category == 'dynamic':
                        ax.plot(rolling_mean, linestyle='--', label='Mean', color='red')

                    # bands (upper, lower)
                    ubs = [rolling_mean]
                    lbs = [rolling_mean]
                    for k in [0.5, 1, 1.5, 2]:
                        ub_suffix = '_ub_' + str(k)
                        lb_suffix = '_lb_' + str(k)
                        ubs.append(ub_arr[method + ub_suffix])
                        lbs.append(lb_arr[method + lb_suffix])

                    for i in range(len(ubs) - 1):
                        lower_ub = ubs[i]
                        upper_ub = ubs[i + 1]

                        upper_lb = lbs[i]
                        lower_lb = lbs[i + 1]

                        ax.fill_between(rolling_mean.index, upper_ub, lower_ub, color='lightgray',
                                        alpha=1 / len(ubs) * i)
                        ax.fill_between(rolling_mean.index, upper_lb, lower_lb, color='#539caf',
                                        alpha=1 / len(lbs) * i)

                ax.set_ylabel(y_label)
                ax.legend(loc='upper left')
                ax.tick_params(labelleft=False, labelright=True, left=False, right=True)
                annotations = [[spread.max(), "H:", "gray"], [spread.min(), "L:", "gray"], [spread[-1], "C:", "blue"]]

                if method_category == 'dynamic' and len(rolling_mean) > 0:
                    annotations = [[rolling_mean[-1], "A:", "red"]] + annotations
                elif method_category == 'total_mean':
                    annotations = [[mean, "A:", "red"]] + annotations
                else:
                    pass

                for annotation in annotations:
                    data, prefix, box_color = annotation[0], annotation[1], annotation[2]
                    bbox_props = dict(boxstyle='larrow', pad=0.01, fc=box_color, ec="b", lw=0)
                    ax.annotate(prefix + '%.3f   ' % data, xy=(1, data), xytext=(0, 0),
                                xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                                bbox=bbox_props, color='white')

        plt.tight_layout()
        plt.close(fig='all')
        return fig, axes

    def spread(self, method, y_label='Spread'):
        fig = plt.figure(figsize=(7, FIG_HEIGHT), dpi=DPI)
        ax = fig.add_subplot(111)

        spread = self.pair.spread[method].loc[self.from_:]
        if len(spread) > 0:
            if method == 'Ratio':
                ax.plot(spread, label='Ratio ({} / {})'.format(self.y_name, self.x_name), color='blue')
            else:
                ax.plot(spread, label=method, color='blue')

            # mean
            mean = self.pair.data[method + '_mean'].dropna().loc[self.from_:]
            if len(mean) > 0:
                if method_category[method] == 'index':
                    ax.plot(mean, linestyle='--', label='Mean', color='red')
                else:
                    ax.axhline(0, color='red', linestyle='--')

                # bands (upper, lower)
                ubs = [mean]
                lbs = [mean]
                for k in [0.5, 1, 1.5, 2]:
                    ub_suffix = '_ub_' + str(k)
                    lb_suffix = '_lb_' + str(k)
                    ubs.append(self.pair.data[method + ub_suffix].loc[self.from_:])
                    lbs.append(self.pair.data[method + lb_suffix].loc[self.from_:])

                for i in range(len(ubs) - 1):
                    lower_ub = ubs[i]
                    upper_ub = ubs[i + 1]

                    upper_lb = lbs[i]
                    lower_lb = lbs[i + 1]

                    ax.fill_between(mean.index, upper_ub, lower_ub, color='lightgray', alpha=1 / len(ubs) * i)
                    ax.fill_between(mean.index, upper_lb, lower_lb, color='#539caf', alpha=1 / len(lbs) * i)

            ax.set_ylabel(y_label)
            ax.legend(loc='upper left')
            ax.tick_params(labelleft=False, labelright=True, left=False, right=True)
            annotations = [[spread.max(), "H:", "gray"], [spread.min(), "L:", "gray"], [spread[-1], "C:", "blue"]]

            if len(mean) > 0:
                annotations = [[mean[-1], "A:", "red"]] + annotations

            for annotation in annotations:
                data, prefix, box_color = annotation[0], annotation[1], annotation[2]
                bbox_props = dict(boxstyle='larrow', pad=0.01, fc=box_color, ec="b", lw=0)
                ax.annotate(prefix + '%.3f   ' % data, xy=(1, data), xytext=(0, 0),
                            xycoords=('axes fraction', 'data'), textcoords='offset points', fontsize=FONT_SIZE,
                            bbox=bbox_props, color='white')

        plt.tight_layout()
        plt.close(fig='all')
        return fig, ax




#from .cybos import CybosPlusStockChart
from .yahoo import YahooEvents, YahooStockChart
from .naver import NaverEvents, NaverStockChart
from datetime import datetime
#from .db import DB
import time

class data:
    def __init__(self):
        pass

    @staticmethod
    def history(asset_arr, field_arr, from_date, to_date=datetime.now().strftime("%Y-%m-%d"), freq="1d"):
        historical_data = {}

        yahoo_stock_chart = YahooStockChart()
        naver_stock_chart = NaverStockChart()

        for asset in asset_arr:
            # if asset.source == 'cybos':
            #     cybos_stock_chart = CybosPlusStockChart()
            #     historical_data[asset.use()] = cybos_stock_chart.request(symbol=asset.use(), fields=field_arr,
            #                                                               from_date=from_date, freq=freq)
            #     time.sleep(0.5)
            if asset.source == 'yahoo':
                symbol = asset.use().replace("/", "-")
                historical_data[asset.use()] = yahoo_stock_chart.request(symbol=symbol, fields=field_arr,
                                                                         from_date=from_date, to_date=to_date,
                                                                         freq=freq)
                time.sleep(0.5)
            # elif asset.source == 'db':
            #     historical_data[asset.use()] = DB(db_fname='ohlcvd.db').request(symbol=asset.use(),
            #                                                                        fields=field_arr,
            #                                                                        from_date=from_date)
            elif asset.source == 'naver':
                historical_data[asset.use()] = naver_stock_chart.request(symbol=asset.use(), fields=field_arr,
                                                                         from_date=from_date, to_date=to_date)

            else:
                raise Exception

        return historical_data

    @staticmethod
    def events(asset_arr, field_arr, start_date="1990-01-01", end_date=datetime.now().strftime("%Y-%m-%d")):
        events_data = {}
        naver_events = NaverEvents()
        yahoo_events = YahooEvents()

        for asset in asset_arr:
            # if asset.source == 'cybos':
            #     events_data[asset.use()] = naver_events.request(symbol=asset.symbol, fields=field_arr)  # 네이버 대체
            #     pass
            #     time.sleep(0.5)  # TR Limit
            if asset.source == 'yahoo':
                symbol = asset.use().replace("/", "-")
                events_data[asset.use()] = yahoo_events.request(symbol=symbol, field_arr=field_arr,
                                                                 from_date=start_date, to_date=end_date)
                time.sleep(0.5)
            elif asset.source == 'naver':
                events_data[asset.use()] = naver_events.request(symbol=asset.use(), fields=field_arr)

            # elif asset.source == 'db':
            #     events_data[asset.use()] = DB(db_fname='cp_dy.db').request_non_time_series(symbol=asset.use(), fields=field_arr)
            else:
                raise Exception
        return events_data


import requests
from bs4 import BeautifulSoup
import pandas as pd
from datetime import datetime
from utils import convert_str2num, find_key_by_val

class NaverEvents:
    def __init__(self):
        self.URL = "https://finance.naver.com/item/main.nhn?code="
        self.fields_dic = {"DIVIDEND_YIELD": "배당수익률", "SHARESOUT": "상장주식수", "PER": "PER", "EPS": "EPS"}

    def request(self, symbol, fields=["DIVIDEND_YIELD", "SHARESOUT"], freq="1d"):
        req_fields = [self.fields_dic[field] for field in fields]
        req_URL = self.URL + str(symbol)
        dataset_raw = requests.get(req_URL)
        soup = BeautifulSoup(dataset_raw.text, 'html.parser')
        response = {}
        try:
            dataset_raw = soup.select('div.tab_con1')[0]

            th_data = []
            th_data_raw = [item.get_text().strip().split('\n')[0] for item in dataset_raw.select('th')]

            for th_raw in th_data_raw:
                if 'l' in th_raw:
                    splitted = th_raw.split('l')
                    th_data.append(splitted[0])
                    if '배당수익률' not in th_raw:
                        th_data.append(splitted[1])
                else:
                    th_data.append(th_raw)

            em_data = [item.get_text().strip().split('\n')[0] for item in dataset_raw.select('em')]

            for key, value in zip(th_data, em_data):
                if key in req_fields:
                    response[find_key_by_val(self.fields_dic, key)] = convert_str2num(value)
        except:
            for field in fields:
                response[field] = 999
        return response
    

class NaverStockChart():
    def __init__(self):
        self.URL = "https://fchart.stock.naver.com/sise.nhn?symbol={}&timeframe=day&count={}&requestType=0"

    def request(self, symbol, from_date='2015-01-01', to_date=datetime.now().strftime("%Y-%m-%d"), fields=["OPEN", "HIGH", "LOW", "CLOSE", "VOLUME"]):
        limit = (datetime.strptime(to_date, '%Y-%m-%d') - datetime.strptime(from_date, '%Y-%m-%d')).days
        req_URL = self.URL.format(str(symbol), limit)

        res = requests.get(req_URL)
        soap = BeautifulSoup(res.text, "html.parser")
        items = soap.find_all("item")

        matrix = []

        for item in items:
            row = item['data'].split('|')
            matrix.append(row)

        df = pd.DataFrame(data=matrix, columns=['DATE', 'OPEN', 'HIGH', 'LOW', 'CLOSE', 'VOLUME'])
        df.set_index(keys='DATE', drop=True, inplace=True)
        df = df[fields]

        df.index = pd.to_datetime(df.index, format="%Y%m%d")
        df = df.sort_index(ascending=True)

        cond1 = df.index >= from_date
        cond2 = df.index <= to_date

        for column in df.columns:
            df[column] = pd.to_numeric(df[column])

        return df.loc[cond1 & cond2, :]

if __name__ == '__main__':
    res = NaverStockChart().request(symbol='005380', to_date="2018-01-01")
    print(res)


import statsmodels.api as sm
from scipy.odr import *
import pandas as pd
import numpy as np
from pykalman import KalmanFilter
from data.source import data
from SETTINGS import K_ARR
from datetime import datetime

class SpreadBuilder():
    def __init__(self, window=365, from_="2015-01-01", to_=datetime.now().strftime("%Y-%m-%d")):
        self.window = window
        self.from_ = from_
        self.to_ = to_

    @staticmethod
    def _diff(y, x):
        spread = y - x
        return spread

    @staticmethod
    def _ratio(y, x):
        spread = y / x
        return spread

    @staticmethod
    def _zscore(y, x, window):
        """
        Ratio Moving Average Differential Method
        """
        ratio = y / x
        rolling_ratio = ratio.rolling(window='{}D'.format(window), min_periods=int(window / 2))
        ratio_ma = rolling_ratio.mean()
        ratio_std = rolling_ratio.std()

        spread = (ratio - ratio_ma) / ratio_std
        return spread.dropna()

    @staticmethod
    def _ROsMA(y, x):
        """
        Ratio Moving Average Differential Method
        """
        ratio = y / x
        ratio_ma = ratio.mean()

        spread = (ratio / ratio_ma - 1) * 100
        return spread.dropna()

    @staticmethod
    def _DROsMA(y, x, window):
        """
        Ratio Moving Average Differential Method
        """
        ratio = y / x
        rolling_ratio = ratio.rolling(window='{}D'.format(window), min_periods=int(window / 2))
        ratio_ma = rolling_ratio.mean()

        spread = (ratio / ratio_ma - 1) * 100
        return spread.dropna()

    @staticmethod
    def _OLS(y, x):
        """
        Ordinary Least Squares Regression Method
        """
        model = sm.OLS(y, sm.tools.add_constant(x)).fit()

        beta = model.params[1]
        intercept = model.params[0]
        spread = y - x * beta - intercept

        return spread, beta, intercept

    @staticmethod
    def f(B, x):
        """
        For TLS
        """
        return B[0] * x + B[1]

    @staticmethod
    def _TLS(y, x):
        """
        Total Least Squares Regression Method
        """
        sx = np.std(x)
        sy = np.std(y)

        linear = Model(SpreadBuilder.f)
        mydata = RealData(x=x, y=y, sx=sx, sy=sy)
        myodr = ODR(mydata, linear, beta0=[5., 2.])
        myoutput = myodr.run()

        beta, intercept = myoutput.beta

        spread = y - x * beta - intercept
        return spread, beta, intercept

    @staticmethod
    def _NPI(y, x):
        y_norm = (y - y.mean()) / y.std()
        x_norm = (x - x.mean()) / x.std()
        spread = y_norm - x_norm

        return spread, y_norm, x_norm

    @staticmethod
    def _zscore(y, x, window):
        """
        Ratio ZScore
        """
        ratio = y / x
        rolling_ratio = ratio.rolling(window='{}D'.format(window), min_periods=int(window / 2))
        ratio_ma = rolling_ratio.mean()
        ratio_std = rolling_ratio.std()

        spread = (ratio - ratio_ma) / ratio_std
        return spread.dropna()

    @staticmethod
    def _KF(y, x, delta=1e-5):
        delta = delta
        trans_cov = delta / (1 - delta) * np.eye(2)

        obs_mat = np.vstack(
            [x, np.ones(x.shape)]
        ).T[:, np.newaxis]

        kf = KalmanFilter(
            n_dim_obs=1,
            n_dim_state=2,
            initial_state_mean=np.zeros(2),
            initial_state_covariance=np.ones((2, 2)),
            transition_matrices=np.eye(2),
            observation_matrices=obs_mat,
            observation_covariance=1.0,
            transition_covariance=trans_cov
        )

        state_means, covs = kf.filter(y)

        beta = state_means[:, 0]
        intercept = state_means[:, 1]

        spread = y - x * beta - intercept

        beta = pd.DataFrame(data=beta, index=spread.index, columns=['Beta'])
        intercept = pd.DataFrame(data=intercept, index=spread.index, columns=['Intercept'])

        return spread, beta, intercept

    @staticmethod
    def build(pair, methods, window):
        # Abs Price
        y = pair.data['y']
        x = pair.data['x']

        # Log Price
        pair.data['log_y'] = log_y = np.log(y)
        pair.data['log_x'] = log_x = np.log(x)

        for method in methods:
            if method == 'diff':
                pair.spread[method] = SpreadBuilder._diff(y=y, x=x)

            elif method == 'Ratio':
                # Abs Price
                pair.spread[method] = SpreadBuilder._ratio(y=y, x=x)

            elif method == 'ROsMA':  # Ratio Moving Average Differential
                # Abs Price
                pair.spread[method] = SpreadBuilder._ROsMA(y=y, x=x)

            elif method == 'DROsMA':  # Ratio Moving Average Differential
                # Abs Price
                pair.spread[method] = SpreadBuilder._DROsMA(y=y, x=x, window=window)

            elif method == 'OLS':  # Ordinary Least Squared
                # Log Price
                pair.spread[method], pair.hedge_ratio[method], pair.intercept[method] = SpreadBuilder._OLS(y=log_y,
                                                                                                           x=log_x)

            elif method == 'TLS':  # Total Least Squared
                # Log Price
                pair.spread[method], pair.hedge_ratio[method], pair.intercept[method] = SpreadBuilder._TLS(y=log_y,
                                                                                                           x=log_x)

            elif method == 'NPI':  # Normalized Price Index Differential
                # Abs Price
                pair.spread[method], pair.data[method + '_y'], pair.data[method + '_x'] = SpreadBuilder._NPI(y=y, x=x)

            elif method == 'KF':  # Kalman Filter
                # Log Price
                pair.spread[method], pair.hedge_ratio[method], pair.intercept[method] = SpreadBuilder._KF(y=log_y,
                                                                                                          x=log_x,
                                                                                                          delta=1e-5)
            elif method == 'zscore':
                pair.spread[method] = SpreadBuilder._zscore(y=y,x=x, window=window)
            else:
                print("Incorrect Spread Method : ", method)
                raise Exception


            rolling_spread = pair.spread[method].rolling(window='{}D'.format(window),
                                                         min_periods=int(window / 2))
            rolling_mean = rolling_spread.mean().dropna()
            rolling_std = rolling_spread.std().dropna()
            pair.data[method + '_mean'] = rolling_mean
            pair.data[method + '_std'] = rolling_std

            for k in K_ARR:
                ub_suffix = '_ub_' + str(k)
                lb_suffix = '_lb_' + str(k)
                pair.data[method + ub_suffix] = rolling_mean + k * rolling_std
                pair.data[method + lb_suffix] = rolling_mean - k * rolling_std

        pair.data['rolling_corr'] = y.pct_change().rolling("{}D".format(window)).corr(x.pct_change())
        return pair

    def calculate(self, pair, methods=['Ratio', 'ROsMA', 'TLS', 'KF'], freq='1d'):
        y_asset = pair.y_asset
        x_asset = pair.x_asset

        prices = data.history([y_asset, x_asset], ["CLOSE"], from_date=self.from_, to_date=self.to_, freq=freq)

        paired_prices = pd.concat([prices[y_asset.use()]["CLOSE"], prices[x_asset.use()]["CLOSE"]], axis=1,
                                  join='inner').dropna()
        paired_prices.columns = ['y', 'x']

        # Abs Price
        pair.data['y'] = paired_prices['y']
        pair.data['x'] = paired_prices['x']

        SpreadBuilder.build(pair=pair, methods=methods, window=self.window)



import statsmodels.api as sm
import statsmodels.tsa.stattools as ts
import numpy as np


class ADF(object):
    def __init__(self):
        self.p_value = None
        self.five_perc_stat = None
        self.perc_stat = None
        self.p_min = .0
        self.p_max = .05
        self.look_back = 63

    def apply_adf(self, time_series):
        model = ts.adfuller(time_series, 1)
        self.p_value = model[1]
        self.five_perc_stat = model[4]['5%']
        self.perc_stat = model[0]

    def use_P(self):
        return (self.p_value > self.p_min) and (self.p_value < self.p_max)

    def use_critical(self):
        return abs(self.perc_stat) > abs(self.five_perc_stat)


class Half_Life(object):
    def __init__(self):
        self.hl_min = 1.0
        self.hl_max = 42.0
        self.look_back = 43
        self.half_life = None

    def apply_half_life(self, time_series):
        lag = np.roll(time_series, 1)
        lag[0] = 0
        ret = time_series - lag
        ret[0] = 0

        # adds intercept terms to X variable for regression
        lag2 = sm.add_constant(lag)

        model = sm.OLS(ret, lag2)
        res = model.fit()

        self.half_life = -np.log(2) / res.params[1]

    def use(self):
        return (self.half_life < self.hl_max) and (self.half_life > self.hl_min)


class Hurst(object):
    def __init__(self):
        self.h_min = 0.0
        self.h_max = 0.4
        self.look_back = 126
        self.lag_max = 100
        self.h_value = None

    def apply_hurst(self, time_series):
        """Returns the Hurst Exponent of the time series vector ts"""
        # Create the range of lag values
        lags = range(2, self.lag_max)

        # Calculate the array of the variances of the lagged differences
        tau = [np.sqrt(np.std(np.subtract(time_series[lag:], time_series[:-lag]))) for lag in lags]

        # Use a linear fit to estimate the Hurst Exponent
        poly = np.polyfit(np.log10(lags), np.log10(tau), 1)

        # Return the Hurst exponent from the polyfit output
        self.h_value = poly[0] * 2.0

    def use(self):
        return (self.h_value < self.h_max) and (self.h_value > self.h_min)





# import os, sys
#
# import sqlite3
# import pandas as pd
# from tqdm import tqdm
# from datetime import datetime
# from more_itertools import unique_everseen
# from calendar import monthrange
# import re
# from utils import get_project_dir, ks_stock_symbols
#
# IN_DIR = os.path.join(get_project_dir(), 'data', 'raw')
# OUT_DIR = os.path.join(get_project_dir(), 'data', 'processed')
#
# class DB:
#     def __init__(self, db_fname):
#         """
#         :param fname: ex) 'ohlcv.db'
#         """
#         self.db_path = os.path.join(OUT_DIR, db_fname)
#
#         if not os.path.isfile(self.db_path):
#             print(" [*] Creating {} ...".format(db_fname))
#
#         self.con = sqlite3.connect(self.db_path)
#         self.cursor = self.con.cursor()
#
#     def request(self, symbol, from_date, fields=["OPEN", "HIGH", "LOW", "CLOSE", "VOLUME"]):
#         req_fields = ["DATES"] + fields
#
#         query = "SELECT {} FROM '{}' WHERE Dates >= '{}'" \
#                 "AND Dates  <= '{}'".format(', '.join(req_fields), symbol, from_date, datetime.now().date())
#
#         self.cursor.execute(query)
#
#         df = pd.DataFrame.from_records(data=self.cursor.fetchall(), columns=req_fields)
#         df = df.set_index('DATES')
#         df = df.apply(pd.to_numeric)
#         df.index = pd.to_datetime(df.index)
#         return df
#
#     def get_table_names(self):
#         query = "SELECT name FROM sqlite_master WHERE type='table';"
#         return [i[0] for i in self._fetch_data(query, fetch_type='all')]
#
#     def get_latest_date(self, symbol):
#         query = "SELECT max(Date) FROM '{}'".format(symbol)
#         return self._fetch_data(query, fetch_type='one')
#
#     def get_oldest_date(self, symbol):
#         query = "SELECT min(Date) FROM '{}'".format(symbol)
#         return self._fetch_data(query, fetch_type='one')
#
#     def _fetch_data(self, query, fetch_type='all'):
#         try:
#             self.cursor.execute(query)
#             if fetch_type == 'all':
#                 return self.cursor.fetchall()
#             elif fetch_type == 'one':
#                 return self.cursor.fetchone()[0]
#         except:
#             return False
#
#     def convert_to_db(self, raw_fname, out_columns=["OPEN", "HIGH", "LOW", "CLOSE", "VOLUME"],
#                       sheet_name='Sheet1', in_type='bloomberg', out_type='default'):
#         """
#         :param raw_fname: 변환할 파일이름
#         :param out_columns: 저장할 컬럼
#         :param sheet_name: 변환할 파일 시트
#         :param in_type: 변환할 파일 타입
#         :param out_type: 출력할 타입
#         """
#         self.in_type = in_type
#         self.out_type = out_type
#
#         raw_path = os.path.join(IN_DIR, raw_fname)
#
#         if self.in_type == 'bloomberg':
#             print(" [*] Converting Bloomberg File to DB File...")
#             self._convert_bloomberg(path=raw_path, sheet_name=sheet_name, out_columns=out_columns)
#
#         elif self.in_type == 'quantiwise':
#             print(" [*] Converting Quantiwise File to DB File...")
#             self._convert_quantiwise(path=raw_path, sheet_name=sheet_name, out_columns=out_columns)
#
#         elif self.in_type == 'non_time_series':
#             print(" [*] Converting Non-Time Series Xlsx File to DB File...")
#             self._convert_non_time_series(path=raw_path, sheet_name=sheet_name, out_columns=out_columns)
#
#         else:
#             raise Exception
#
#     def request_non_time_series(self, symbol, fields=["DIVIDEND_YIELD"]):
#         query = "SELECT {} FROM '{}'".format(', '.join(fields), symbol)
#         self.cursor.execute(query)
#         df = pd.DataFrame.from_records(data=self.cursor.fetchall(), columns=fields).iloc[0, :].to_dict()
#         return df
#
#     def _convert_non_time_series(self, path, sheet_name, out_columns):
#         # settings
#         data = pd.read_excel(path, sheet_name=sheet_name)
#         symbols = data.columns
#
#         with tqdm(total=len(symbols), file=sys.stdout) as pbar:
#             for i, symbol in enumerate(symbols):
#                 df = data[symbol]
#
#                 df = df.to_frame().T
#                 df.columns = out_columns
#
#                 if len(df) == 0:
#                     print(df)
#                     pass
#                 df.to_sql(symbol, con=self.con, if_exists='replace', index=False)
#                 pbar.update(1)
#
#     def _convert_bloomberg(self, path, sheet_name, out_columns):
#         # settings
#         n_features = len(out_columns)
#         symbols_idx = 0
#         features_idx = 1
#         data = pd.read_excel(path, sheet_name=sheet_name, header=None)
#
#         # get symbol lists
#         symbols = [data.ix[symbols_idx][i] for i in range(1, len(data.ix[symbols_idx]), n_features)]
#         data = pd.DataFrame(data.ix[features_idx + 1:, ])
#         data = data.set_index(0)
#         data.index.name = 'DATES'
#
#         k = 0
#         with tqdm(total=len(symbols), file=sys.stdout) as pbar:
#             for i in range(len(symbols)):
#                 symbol = symbols[i]
#                 df = pd.DataFrame(data.ix[:, i * n_features + k:i * n_features + n_features].values, index=data.index,
#                                   columns=out_columns)
#                 k = 1
#                 df.index = pd.to_datetime(df.index)
#                 df = df.dropna(how='all')
#                 if len(df) == 0:
#                     print(df)
#                     pass
#                 df.to_sql(symbol, con=self.con, if_exists='replace')
#                 pbar.update(1)
#
#     def _convert_quantiwise(self, path, sheet_name, out_columns):
#         """
#         :param data_path: raw data path
#         :param sheet_type: single, multiple
#         :return:
#         """
#         with pd.ExcelFile(path) as reader:
#             sheet = pd.read_excel(reader, sheet_name=sheet_name)
#             total_df, symbols = self._quantisheet_to_df(sheet, out_columns)
#
#             with tqdm(total=len(symbols), file=sys.stdout) as pbar:
#                 for symbol in symbols:
#                     df = total_df[symbol].dropna(how='all')
#                     df.to_sql(symbol, con=self.con, if_exists='replace')
#                     pbar.update(1)
#
#     def _quantisheet_to_df(self, sheet, out_columns):
#         symbols = list(unique_everseen(sheet.iloc[6, :]))
#         symbols.remove('Code')
#
#         converted_symbols = ks_stock_symbols(symbols, in_type=self.in_type, out_type=self.out_type)
#
#         sheet.columns = sheet.iloc[12, :]
#         sheet = sheet.iloc[13:, ]
#         sheet = sheet.set_index('D A T E')
#         sheet.index.name = 'DATES'
#
#         df = sheet.iloc[:-1, :]
#         df.columns = pd.MultiIndex.from_product([converted_symbols, out_columns])
#
#         return df, converted_symbols
#
#     def close(self):
#         self.con.close()
#
# def main():
#
#     DB(db_fname='ohlcvd.db').convert_to_db(raw_fname='CEF_ETF_ohlcvd.xlsx',
#                                                        out_columns=["OPEN", "HIGH", "LOW", "CLOSE", "VOLUME", "DIVIDEND_YIELD"],
#                                                        sheet_name='Sheet1',
#                                                        in_type='bloomberg',
#                                                        out_type='default')
#
#     """
#     print(DB(db_fname='cp_ohlcv.db').convert_to_db(raw_fname='bloom_cp_dy.xlsx', out_columns=["DIVIDEND_YIELD"],
#                       sheet_name='Sheet1', in_type='non_time_series', out_type='default'))
#     """
#     #print(DB(db_fname='cp_dy.db').request_non_time_series("VLA FP Equity", fields=["DIVIDEND_YIELD", "TEST"]))
#
# if __name__ == '__main__':
#     main()


index_dic = {'KOSPI': "001", "KOSDAQ": "201", "MSCI": "530", "KOSPI200": "180"}

# Dont Forget '.'
suffix_yahoo_dic = {'US':'', 'IM':'.MI', 'DC':'.CO', 'FH':'.HE', 'BZ':'.SA', 'RM':'.ME', 'CI':'.SN',
                    'AR':'.BA', 'CN':'.TO', 'SS':'.ST', 'LN':'.L', 'MM':'.MX', 'NA':'.AS', 'SM':'.MC', 'NO':'.OL',
                    'CH':'.SS', 'GR':'.DE', 'CH':'.SS'}
st_yahoo_dic = ['SCA']


# TODO: SS:.ST? SS:.SW?

prefix_cybos_dic = {'stock':'A', 'etf':'A', 'index':'U'}

prefix_dic = {'cybos':prefix_cybos_dic}
suffix_dic = {'yahoo':suffix_yahoo_dic}

def suffix_factory(source, exchange):
    if source in suffix_dic.keys():
        if exchange in suffix_dic[source].keys():
            return '{}'.format(suffix_dic[source][exchange])
        else:
            if source == 'yahoo':
                return '.{}'.format(exchange)
    else:
        if source == 'bloomberg':
            return ' {} Equity'.format(exchange)
        else:
            return ''

class Asset:
    def __init__(self, symbol, type_, exchange='KS', source='cybos', name=None, **kwargs):
        self.symbol, self.type, self.exchange, self.source = symbol, type_, exchange, source
        self.data = {}

        if name is None:
            self.name = symbol
        else:
            self.name = name

        # Add additonal data when to initalize (Optional)
        for key, value in kwargs.items():
            exec("self.data['{}'] = '{}'".format(key, value))

        self.suffix = suffix_factory(self.source, self.exchange)
        self.prefix = prefix_dic[self.source][self.type] if self.source in prefix_dic.keys() else ''

    def use(self):
        return self.prefix + self.symbol + self.suffix

def asset_wrapper(type_):
    class ProductClass(Asset):
        def __init__(self, symbol, exchange='KS', source='cybos', name=None, **kwargs):
            if type_ == 'index' and source=='cybos':
                symbol = index_dic[symbol] if symbol in index_dic.keys() else symbol
            super().__init__(symbol=symbol, type_=type_, exchange=exchange, source=source, name=name)
            for key, value in kwargs.items():
                exec("self.data['{}'] = '{}'".format(key, value))
    return ProductClass


def asset_factory(symbol, name=None, type_='stock', exchange='KS', source='cybos'):
    if type_ == 'stock':
        return Stock(symbol=symbol, name=name, exchange=exchange, source=source)
    elif type_ == 'etf':
        return ETF(symbol=symbol, name=name, exchange=exchange, source=source)
    elif type_ == 'index':
        return Index(symbol=symbol, name=name, exchange=exchange, source=source)
    else:
        return Asset(symbol=symbol, type_=type_, name=name, exchange=exchange, source=source)


class Pair:
    def __init__(self, y_asset, x_asset, **kwargs):
        self.y_asset = y_asset
        self.x_asset = x_asset

        self.data = {}
        self.spread = {}
        self.hedge_ratio = {}
        self.intercept = {}

        # Add additonal data when to initalize (Optional)
        for key, value in kwargs.items():
            exec("self.data['{}'] = '{}'".format(key, value))

Stock = asset_wrapper(type_='stock')
ETF = asset_wrapper(type_='etf')
Index = asset_wrapper(type_='index')

def main():
    stock = asset_factory(symbol='KOSPI', type_='index', source='cybos', exchange='KS')
    print(stock.use())

if __name__ == '__main__':
    main()





import os, sys
project_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
sys.path.insert(0,project_dir)

from pathlib import Path
from builds.asset import asset_factory, Pair
from data.source import data, NaverEvents
from builds.xlsx_manager import XlsxMgr
from utils import get_project_dir, bloomberg2asset
from builds.pairplotlib import PairPlotLib
from models.spread_builder import SpreadBuilder
from builds.statistics import ADF, Half_Life, Hurst
import itertools

import numpy as np
import pandas as pd

import warnings
warnings.filterwarnings("ignore")

from tqdm import tqdm
from datetime import datetime, timedelta
from SETTINGS import METHODS, FROM_, PLOT_FROM_, ROLLING, WINDOW, ADDITIONAL_FEATURES, SHEET_COLUMNS, SHEET_HEIGHT

IN_DIR = os.path.join(get_project_dir(), 'data', 'peer_groups')

class PeerGroup:
    settings = {
        'window': 365,
        'analysis': ("2015-01-01", datetime.now().strftime("%Y-%m-%d")),
        'plot': ('2015-01-01', datetime.now().strftime("%Y-%m-%d")),
    }

    def __init__(self, params={}):
        self.settings.update(params)
        self._analysis = self._summary = []

    def set_peer_group(self, peer_group, **kwargs):
        lst = peer_group[['LONG ASSET SYMBOL', 'LONG ASSET NAME', 'SHORT ASSET SYMBOL', 'SHORT ASSET NAME']].values.tolist()
        lst = [[[x[0], x[1]], [x[2], x[3]]] for x in lst]

        self._group = lst

    def analyze(self, filter_ks=True):
        for pair in tqdm(self._group, total=len(self._group), desc=" [*] Analyzing Pairs..."):
            y_asset = bloomberg2asset(bloomberg_symbol=pair[0][0], name=pair[0][1], filter_ks=filter_ks)
            x_asset = bloomberg2asset(bloomberg_symbol=pair[1][0], name=pair[1][1], filter_ks=filter_ks)

            pair = Pair(y_asset=y_asset, x_asset=x_asset)

            try:
                spread_builder = SpreadBuilder(window=self.settings['window'], from_=self.settings['analysis'][0])
                if 'TLS' not in METHODS:  # TLS Scatter PLOT
                    spread_builder.calculate(pair, methods=['TLS'] + METHODS)
                else:
                    spread_builder.calculate(pair, methods=METHODS)
            except Exception as e:
                print(e)
                continue

            self._add_features(pair, ADDITIONAL_FEATURES)
            self._add_p_statistics(pair, METHODS)
            self._analysis.append(pair)

    def __len__(self):
        return len(self._group)

    def __getitem__(self, position):
        return self._group[position]

    def _add_p_statistics(self, pair, methods):
        for method in methods:
            adf = ADF()
            half_life = Half_Life()
            hurst = Hurst()
            spread = pair.spread[method].dropna()
            try:
                adf.apply_adf(spread)
            except:
                adf.p_value = 999
            try:
                half_life.apply_half_life(spread)
            except:
                half_life.half_life = 999
            try:
                hurst.apply_hurst(spread)
            except:
                hurst.h_value = 999

            pair.data[method + '_p_value'] = adf.p_value
            pair.data[method + '_half_life'] = half_life.half_life
            pair.data[method + '_hurst'] = hurst.h_value

    def _add_features(self, pair, features):
        for feature in features:
            if feature == 'AVG_TRADING_VALUE':
                close_volume_dic = data.history([pair.x_asset, pair.y_asset],
                                                field_arr=["CLOSE", "VOLUME"], from_date=(datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d'))
                for asset in [pair.x_asset, pair.y_asset]:
                    close_volume = close_volume_dic[asset.use()]
                    asset.data[feature] = np.mean(close_volume.iloc[:, 0] * close_volume.iloc[:, 1])

            elif feature == 'DIVIDEND_YIELD':
                # 배당 데이터 받기
                start_date = datetime.now() - pd.DateOffset(days=365)

                for i, asset in enumerate([pair.y_asset, pair.x_asset]):
                    try:
                        if asset.exchange == 'KS':
                            dy = NaverEvents().request(symbol=asset.symbol, fields=[feature])[feature] / 100
                        else:

                            div = data.events(
                                [asset],
                                field_arr=["DIVIDENDS"],
                                start_date=self.settings['analysis'][0])

                            dvd = div[asset.use()]['DIVIDENDS']
                            dvd.sort_index(ascending=False)
                            price = pair.data['y'][-1] if i == 0 else pair.data['x'][-1]
                            dy = dvd.loc[start_date:].values.sum() / price
                        asset.data[feature] = dy
                    except:
                        asset.data[feature] = -999

    def summarize(self, outpath=os.path.join(get_project_dir(), 'output', 'peer_analysis',
                                             '{}'.format(datetime.now().strftime('%Y%m%d%H%M%S')))):

        Path(outpath).mkdir(exist_ok=True)
        xlsx_manager = XlsxMgr(path=outpath, fname='summary.xlsx')

        for i, method in enumerate(METHODS):

            xlsx_manager.add_sheet(method)
            xlsx_manager.set_sheet(method)

            # Initialize Sheet
            for col, items in enumerate(SHEET_COLUMNS.items()):
                key, value = items[0], items[1]
                xlsx_manager.sheet.set_column(col, col, float(value[1]), xlsx_manager.format_dic[value[2]])
                xlsx_manager.write(0, col, value[0], cell_format=xlsx_manager.format_dic['header'])
                xlsx_manager.sheet.write_comment(row=0, col=col, comment=value[3])

            matrix = np.empty(shape=(len(self._analysis), len(SHEET_COLUMNS)), dtype=object)
            for row, pair in tqdm(enumerate(self._analysis), total=len(self._analysis),
                                    desc=" {} Method ... ({}/{})".format(method, i+1, len(METHODS))):

                ppl = PairPlotLib(pair, from_=self.settings['plot'][0])
                xlsx_manager.sheet.set_row(row + 1, SHEET_HEIGHT)

                for col, items in enumerate(SHEET_COLUMNS.items()):
                    key, format_ = items[0], xlsx_manager.format_dic[items[1][2]]
                    if key == 0: # 순서
                        value = row + 1
                    elif key == 1:
                        value = '{} {} Equity'.format(pair.y_asset.symbol, pair.y_asset.exchange)
                    elif key == 2:
                        value = '{} {} Equity'.format(pair.x_asset.symbol, pair.x_asset.exchange)
                    elif key == 3:
                        value = pair.y_asset.name
                    elif key == 4:
                        value = pair.x_asset.name
                    elif key == 5:
                        value = pair.spread[method][-1] if len(pair.spread[method]) > 0 else None
                    elif 6 <= key <= 10:
                        value = ' '

                        if key == 6:
                            y = pair.data['y']
                            x = pair.data['x']
                            fig, _ = ppl.price(x=x, y=y, title='Abs Price')
                        elif key == 7:
                            fig, _ = ppl.spread(method=method)

                        elif key == 8:
                            fig, _ = ppl.spread_box(method=method)

                        elif key == 9:
                            fig, _ = ppl.spread_dist(method=method)

                        elif key == 10:
                            fig, _ = ppl.linear_regression(method='TLS')
                        fig.savefig(xlsx_manager.get_path(path=outpath,
                                                          file_name='{}_{}_{}_{}.png'.format(
                                                              pair.x_asset.symbol,
                                                              pair.y_asset.symbol,
                                                              method, key),
                                                          is_img=True), bbox_inches="tight")
                        xlsx_manager.insert_img(row + 1, col,
                                                file_name='{}_{}_{}_{}.png'.format(pair.x_asset.symbol,
                                                                                   pair.y_asset.symbol,
                                                                                   method, key),
                                                scale={'y_offset': 15, 'x_offset': 15})
                    elif key == 11:
                        value = pair.data['y'][-1]
                    elif key == 12:
                        value = pair.data['x'][-1]
                    elif key == 13:
                        value = pair.data['rolling_corr'][-1] if len( pair.data['rolling_corr']) > 0 else None
                    elif key == 14:
                        value = pair.data[method + '_p_value']
                    elif key == 15:
                        value = pair.data[method + '_hurst']
                    elif key == 16:
                        value = pair.data[method + '_half_life']
                    elif key == 17:
                        value = pair.y_asset.data['AVG_TRADING_VALUE']
                    elif key == 18:
                        value = pair.x_asset.data['AVG_TRADING_VALUE']
                    elif key == 19:
                        value = pair.y_asset.data["DIVIDEND_YIELD"]
                    elif key == 20:
                        value = pair.x_asset.data["DIVIDEND_YIELD"]
                    else:
                        raise Exception

                    xlsx_manager.write(row+1, col, value, cell_format=format_)
                    matrix[row, col] = value

            xlsx_manager.sheet.autofilter(0, 0, 0, len(SHEET_COLUMNS) - 1)

        writer = xlsx_manager.writer
        xlsx_manager.close()

        return

def main():
    cred_path = r'C:\Users\Administrator\OneDrive\Code Archieves\pairs-framework\data\_cridentials\spreadsheet-245601-24ecd3f4c25f.json'
    peer_group = PeerGroup(cred_path=cred_path)
    peer_group.set_peer_group(sheet_name='pairs', **{'LONG ASSET SYMBOL': '000075 KS Equity'})
    peer_group.analyze(filter_ks=True)
    peer_group.summarize(out_dirname='pfd_pfd_ks_')

if __name__ == '__main__':
    main()



import os
import pandas as pd
from pathlib import Path
from utils import is_inf, is_nan
import numpy as np
from utils import get_project_dir

class XlsxMgr():
    def __init__(self, path, fname):
        self.path = path
        self.fname = fname

        self.img_path = os.path.join(self.path, 'imgs')
        Path(self.img_path).mkdir(exist_ok=True)
        self.writer = pd.ExcelWriter(os.path.join(self.path, self.fname), engine='xlsxwriter')

        self.workbook = self.writer.book
        self.sheet_dic = {}
        self.format_dic = self.set_formats()

        self.sheet = None
        self.sheet_name = None

    def add_sheet(self, sheet_name):
        self.sheet_dic[sheet_name] = self.workbook.add_worksheet(sheet_name)

    def set_sheet(self, sheet_name):
        self.sheet = self.sheet_dic[sheet_name]
        self.sheet_name = sheet_name

    def set_formats(self):
        def wrap_dic(dic):
            ret = {'font_name': 'Calibri', 'font_size': 9}
            ret.update(dic)
            return ret

        integer = self.workbook.add_format(wrap_dic({'num_format': '#,##0'}))
        float_second_dec = self.workbook.add_format(wrap_dic({'num_format': '#,##0.00'}))
        float_third_dec = self.workbook.add_format(wrap_dic({'num_format': '#,##0.000'}))
        float_rate = self.workbook.add_format(wrap_dic({'num_format': '0.00%'}))

        date = self.workbook.add_format(wrap_dic({'num_format': 'yyyy-mm-dd'}))

        string = self.workbook.add_format(wrap_dic({'font_name': 'Calibri'}))
        string_bold_red = self.workbook.add_format(wrap_dic({'font_color': 'red', 'bold': 1}))
        string_bold_blue = self.workbook.add_format(wrap_dic({'font_color': 'blue', 'bold': 1}))
        string_bold = self.workbook.add_format(wrap_dic({'bold': 1}))

        header = self.workbook.add_format(wrap_dic({'font_color': 'white',
                                                    'text_wrap': True,
                                                    'fg_color': '#084B8A',
                                                    'border': 0}))
        numeric = self.workbook.add_format(wrap_dic({'border': 0}))

        format_dic = {"float_second_dec": float_second_dec,
                      "float_third_dec": float_third_dec,
                      "string": string,
                      "date": date,
                      "string_bold_red": string_bold_red,
                      "string_bold_blue": string_bold_blue,
                      "float_rate": float_rate,
                      "string_bold": string_bold,
                      "integer": integer,
                      "header": header,
                      "numeric": numeric
                      }

        for key, value in format_dic.items():
            eval(key + '.set_align("center")')
            eval(key + '.set_align("vcenter")')

        return format_dic

    def write(self, col, row, data, cell_format):
        if not (data is None or is_inf(data) or is_nan(data)):  # nan, inf, None check
            self.sheet.write(col, row, data, cell_format)
        else:
            pass

    def df_to_excel(self, df, sheet_name):
        df.to_excel(self.writer, sheet_name=sheet_name, startrow=0, index=False)
        self.sheet = self.writer.sheets[sheet_name]

    def insert_img(self, col, row, file_name, scale=None):
        self.sheet.insert_image(col, row, self.get_path(path=self.path, file_name=file_name, is_img=True), scale)

    def close(self):
        # self.sheet.autofilter(0, 0, 0, len(cols)-1)
        self.writer.save()

    @staticmethod
    def get_path(path, file_name, is_img=False):
        # If '/' in name, convert it to '_'
        converted_name = file_name.replace('/', '_')
        if is_img == False:
            path = os.path.join(path, converted_name)
        else:
            path = os.path.join(path, 'imgs', converted_name)
        return path

    def init_sheet(self, n_rows, cols_meta, row_height=230):
        # Sheet Column Name, Width, format, Df Column, Dtype(-1- x , 0- x+y, 1- y)

        for i, key, value in enumerate(cols_meta.items()):
            self.sheet.set_column(i, i, value[0], cell_format=cols_meta[value[1]])
            self.sheet.write(0, i, key, self.format_dic['header'])
            self.sheet.write_comment(0, i, value[2])

        for i in range(n_rows):
            self.sheet.set_row(i + 1, row_height)





import os
import math
from builds.asset import asset_factory
import pandas as pd

def bloomberg2asset(bloomberg_symbol, name=None, filter_ks=True):
    str = bloomberg_symbol.split()
    symbol, exchange = str[0], str[1]
    source = 'naver' if exchange == 'KS' and filter_ks == True else 'yahoo'
    symbol = symbol.replace("/", "-")
    asset = asset_factory(symbol=symbol, exchange=exchange, source=source) if name is None \
        else asset_factory(symbol=symbol, exchange=exchange, source=source, name=name)
    return asset


def extract_years(df):
    current_year = int(df.index[0].strftime("%Y"))
    last_year = int(df.index[-1].strftime("%Y"))

    years = []

    while current_year <= last_year:
        years.append(current_year)
        current_year += 1

    return years

def split_yearly(df, method='ffill'):
    years = extract_years(df)

    date_range = pd.date_range('1999-01-01', '1999-12-31')
    yearly_data = pd.DataFrame(index=[date.strftime("%m-%d") for date in date_range])

    for year in years:
        single_year_data = df.loc['{}-01-01'.format(year): '{}-01-01'.format(year + 1)]
        for idx, row in single_year_data.iteritems():
            yearly_data.loc[yearly_data.index == idx.strftime("%m-%d"), str(year)] = row

    return yearly_data.fillna(method=method)

def convert_str2num(string):
    try:
        return float(string.replace(',', ''))
    except:
        if '%' in string:
            return float(string.strip('%')) / 100
        else:
            return string

def find_key_by_val(dic, val):
    for key, value in dic.items():
        if val == value:
            return key
    return "key doesn't exist"

def get_project_dir():
    return os.path.dirname(os.path.realpath(__file__))

def ks_stock_symbols(symbols, in_type='default', out_type='bloomberg'):
    """
    Symbol을 Bloomberg 혹은 Quantiwise(Cybos) Symbol로 변환
    """
    if in_type == 'default':
        if out_type == 'padding': # Zero padding
            for i, code in enumerate(symbols):
                n_str = len(str(code))
                for _ in range(6-n_str):
                    symbols[i] = "0" + symbols[i]
            return symbols
    elif in_type == 'bloomberg':
        symbols = [symbol[:6] for symbol in symbols]

    elif in_type == 'quantiwise':
        symbols = [symbol[1:] for symbol in symbols]

    else:
        raise Exception

    if out_type == 'default':
        pass

    elif out_type == 'bloomberg':
        symbols = ['{} KS Equity'.format(symbol[:]) for symbol in symbols]

    elif out_type == 'quantiwise':
        symbols = ['A{}'.format(symbol[:]) for symbol in symbols]

    else:
        raise Exception

    return symbols

def is_nan(x):
    return isinstance(x, float) and math.isnan(x)

def is_inf(x):
    return isinstance(x, float) and math.isinf(x)
